// 
// Microsoft Windows XP / Vista / 7 / 8 Win32k.sys Syscall Fuzzer
//
// Author: Mateusz "j00ru" Jurczyk
// Date:   2010-06-16, refactored 2012-12-08
// E-Mail: j00ru.vx@gmail.com
// WWW:    http://j00ru.vexillium.org/
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <Dpfilter.h>
#include <Windows.h>


// Native API definitions
ULONG __cdecl DbgPrintEx(
  ULONG ComponentId,
  ULONG Level,
  PCSTR Format,
  ...
);

VOID DbgBreakPoint();

// Constant base identifier for syscalls of a specific subsystem (win32k.sys here)
#define kSystemServiceBase 0x1000

// Maximum number of parameters a win32k.sys system call can take. The current value
// is a safe bet; however, it is possible that service with more arguments exist.
#define kMaxSyscallParams 7

// Number of win32k.sys syscalls on the supported platform
#define kSyscallCount 0x33c

// Win32k.sys service id numbers excluded from fuzzing.
BOOLEAN DisabledAPI[0x2000];

// Currently fuzzed win32k.sys system call id.
ULONG CurrentApiNumber = kSystemServiceBase;

// Current fuzzing iteration (used to detect hang threads).
ULONG Iteration;

// RDTSC-based PRNG routine.
ULONGLONG (*__rdtsc_random)() = (ULONGLONG (*)())"\x0F\x31"         // RDTSC
                                                 "\x89\xC1"         // MOV ECX, EAX
                                                 "\x89\xD0"         // MOV EAX, EDX
                                                 "\x48\xC1\xE0\x20" // SHL RAX, 32
                                                 "\x48\x01\xC8"     // ADD RAX, RCX
                                                 "\x48\xD1\xE8"     // SHR RAX, 1
                                                 "\xC3";            // RET

// Assembly helper to directly invoke win32k.sys system calls
UCHAR SyscallStubBytes[] = "\x4C\x8B\xD1"         // MOV R10, RCX
                           "\xB8\x00\x00\x00\x00" // MOV EAX, <SYSCALL NUMBER>
                           "\x0F\x05"             // SYSCALL
                           "\xC3";                // RET
ULONG (*SyscallWrapper)() = (ULONG (*)())SyscallStubBytes;

ULONG InvokeService(ULONG ServiceId, ULONGLONG a, ULONGLONG b, ULONGLONG c, ULONGLONG d, ULONGLONG e,
                    ULONGLONG f, ULONGLONG g) {
  PUCHAR SyscallWrapperBytes = (PUCHAR)SyscallWrapper;
  memcpy(&SyscallWrapperBytes[4], &ServiceId, sizeof(ULONG));

  // Invoke the system call.
  return SyscallWrapper(a, b, c, d, e, f, g);
}

ULONG DebugPrint(PCHAR Format, ...) {
  CHAR szMessageBuffer[1024];
  va_list va;

  va_start(va, Format);
  _vsnprintf(szMessageBuffer, sizeof(szMessageBuffer), Format, va);
  va_end(va);

  return DbgPrintEx(DPFLTR_DRIVEEXTENDER_ID, DPFLTR_MASK | DPFLTR_INFO_LEVEL, szMessageBuffer);
}

DWORD FuzzingThread(LPVOID lpParameter) {
  DWORD OldProtect;

  // Make sure snippets of inline assembly are executable
  VirtualProtect(SyscallWrapper, sizeof(SyscallWrapper), PAGE_EXECUTE_READWRITE, &OldProtect);
  VirtualProtect(__rdtsc_random, sizeof(__rdtsc_random), PAGE_EXECUTE_READWRITE, &OldProtect);

  // Convert the thread to a GUI one.
  LoadLibraryA("user32.dll");

  // Loop indefinitely
  while(1) {
    ULONGLONG dwParams[kMaxSyscallParams];
    ULONG Return;
    ULONG i;

    // Pick the system call id for this iteration.
    CurrentApiNumber = kSystemServiceBase + (__rdtsc_random() % kSyscallCount);

    // If the service has been excluded for whatever reason, skip it
    if (DisabledAPI[CurrentApiNumber]) {
      continue;
    }

    // Randomly choose syscall parameters using an exceedingly simple algorithm.
    for (i = 0; i < kMaxSyscallParams; i++) {
      if (__rdtsc_random() & 1) {
        dwParams[i] = __rdtsc_random();
      } else {
        dwParams[i] = 0;
      }
    }

    // Print out information about the current iteration for logging purposes.
    DebugPrint("CurrentApiNumber = 0x%#x\nParams: ", CurrentApiNumber);
    for (i = 0; i < kMaxSyscallParams; i++) {
      DebugPrint("%#x ", dwParams[i]);
    }
    DebugPrint("\n");

    // Invoke the win32k.sys service with a variable number of parameters specified
    // at compile time.
    Return = InvokeService(CurrentApiNumber, dwParams[0], dwParams[1], dwParams[2], dwParams[3],
                           dwParams[4], dwParams[5], dwParams[6]);

    // Increment current iteration number.
    Iteration++;
  }
  return 0;
}

DWORD WatchdogThread(LPVOID lpParameter) {
  ULONG OldIteration;

  // Create an initial fuzzing thread.
  HANDLE hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)FuzzingThread, NULL, 0, NULL);

  // Reset counters.
  OldIteration = Iteration = 0;

  // Loop indefinitely.
  while(1) {
    OldIteration = Iteration;

    // Sleep for a second.
    Sleep(1000);

    // If the fuzzing thread has been stuck on a particular syscall for ~1 second, restart it
    // and exclude the service from future use.
    if(OldIteration == Iteration)
    {
      DisabledAPI[CurrentApiNumber] = TRUE;

      TerminateThread(hThread, 0);
      hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)FuzzingThread, NULL, 0, NULL);
    }
  }
}

LONG CALLBACK VectoredHandler(PEXCEPTION_POINTERS ExceptionInfo) {
  // When an exception occurs, terminate the current thread gracefully,
  // by redirecting it into kernel32!ExitThread
  DWORD64 ExitThread = (DWORD64)GetProcAddress(GetModuleHandle("kernel32.dll"), "ExitThread");
  ExceptionInfo->ContextRecord->Rip = ExitThread;

  return EXCEPTION_CONTINUE_EXECUTION;
}

BOOL WINAPI DllMain(HANDLE hinstDLL, DWORD dwReason, LPVOID lpvReserved) {
  if(dwReason == DLL_PROCESS_ATTACH) {
    OSVERSIONINFOEX osvi;

    DebugPrint("[+] In syscall_fuzz!DllMain\n");

    // Make sure we're executing on Windows 7 64-bit.
    ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
    if (!GetVersionEx((OSVERSIONINFO*) &osvi)) {
      DebugPrint("[-] GetVersionEx failed, %u\n", GetLastError());
      return FALSE;
    }
    if (osvi.dwMajorVersion != 6 || osvi.dwMinorVersion != 1) {
      DebugPrint("[-] System is not Microsoft Windows 7 64-bit\n");
      return FALSE;
    }

    // Create a watchdog thread on attach
    CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WatchdogThread, NULL, 0, NULL);

    // Add an exception handler in case something goes wrong (e.g. a crash in a user-mode
    // callback).
    AddVectoredExceptionHandler(1, VectoredHandler);
  }

  return TRUE;
}

